<?php

namespace Goldenscarab\Modulus\Debusca\App\Http\Drivers;

use PDO;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Cache;
use Goldenscarab\Modulus\Debusca\App\Http\Drivers\Objects\Meta;
use Goldenscarab\Modulus\Debusca\App\Http\Drivers\Objects\MetaCollect;
use Goldenscarab\Modulus\Debusca\App\Http\Drivers\Objects\QueryResult;


class Sqlsrv extends Driver
{
    private $queries = [
        'tables'  => "SELECT * FROM information_schema.tables ORDER BY TABLE_NAME ASC",
        'schema'  => "SELECT object_name(object_id) as table_name, type_name(system_type_id) as type, object_definition(default_object_id) as default_value, * FROM sys.columns;",
        'select'  => "SELECT TOP 25 * FROM {TABLE}",
        'insert'  => "INSERT INTO `{TABLE}` ({COLUMNS_NAMES}) VALUES ({COLUMNS_VALUES});",
        'update'  => "UPDATE `{TABLE}` AS {ALIAS} SET {NAMES_VALUES} WHERE {ALIAS}.`{PK_NAME}` = {ID};",
        'delete'  => "DELETE FROM `{TABLE}` WHERE `{PK_NAME}` = {ID};",
        'create'  => "CREATE TABLE `my_table` (`id` INT(10) NOT NULL, `name` VARCHAR(255) NOT NULL, `status` TINYINT(1) NOT NULL DEFAULT '0', `date` DATE NULL, PRIMARY KEY (`id`)) ENGINE = InnoDB;",
        'alter'   => "ALTER TABLE `{TABLE}` ADD `[COL_NAME]` varchar(255) NOT NULL AFTER [OTHER_COL];\nALTER TABLE `{TABLE}` CHANGE `[COL_OLD_NAME]` `[COL_NEW_NAME]` VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_unicode_ci NOT NULL;\nALTER TABLE `{TABLE}` MODIFY COLUMN [COL_NAME] varchar(255) NOT NULL AFTER [OTHER_COL_NAME];"
    ];

    private $tables_metas = null;


    public function getTables(): Collection
    {
        $cache_key = 'sqlsrv.' . $this->db->getConfig('name') . '.tables';

        $tables = Cache::remember($cache_key, 5 * 60, function () {

            $tables = collect();

            $result = $this->db->select($this->queries['tables']);
            $tables = collect($result)->map(function($item) {
                return data_get($item, 'TABLE_NAME');
            });

            $collect = collect();

            foreach ($tables as $table) {
                $metas = $this->getMetas($table);
                $qr = new QueryResult($metas, [$table], $this->queries);
                $collect->push($qr);
            }

            return $collect;
        });

        return $tables;
    }

    public function getMetas($table): Collection
    {
        $cache_key = 'sqlsrv.' . $this->db->getConfig('name') . '.metas';

        $tables_metas = &$this->tables_metas;

        $metas = Cache::remember($cache_key, 5 * 60, function () use ($table, $tables_metas) {

            if (is_null($tables_metas)) {
                $tables_metas = collect($this->db->select($this->queries['schema']));
            }

            $columns = $tables_metas->where('table_name', $table);

            $metas = new MetaCollect();

            foreach ($columns as $item) {

                $meta = new Meta;
                $meta->setTable($table);
                $meta->setField($item->name);
                $meta->setType($item->type);
                $meta->setNull($item->is_nullable);
                $meta->setKey($item->is_identity ? 'primary_key' : '');
                $meta->setDefault($item->default_value);
                //$meta->setExtra(null);

                $metas->push($meta);
            }

            return $metas;
        });

        return $metas;
    }

    public function getSuggestions(): Collection
    {
        $tables_metas = &$this->tables_metas;
        $cache_key = 'sqlsrv.' . $this->db->getConfig('name') . '.suggestions';

        $suggestions = Cache::remember($cache_key, 5 * 60, function () use ($tables_metas) {

            $suggestions = collect();

            if (is_null($tables_metas)) {
                $tables_metas = $this->db->select($this->queries['schema']);
            }

            foreach ($tables_metas as $column) {
                $suggestions->push($column->table_name);
                $suggestions->push($column->name);
                $suggestions->push($column->table_name . '.' . $column->name);
            }

            return $suggestions;
        });

        return $suggestions;
    }

    public function query($query): QueryResult
    {
        $time_start = microtime(true);
        $splited = explode(' ', strtolower($query));
        $type    = array_shift($splited);
        $result  = null;

        $_query = DB::raw($query);
        $statement = $this->db->getPdo()->prepare($_query);
        $status = $statement->execute();
//dd($statement);
        switch ($type) {
            case 'select':
            case 'describe':
            case 'show':
            case 'exec':
                $result = $statement->fetchAll(PDO::FETCH_ASSOC);
                break;
            case 'update':
            case 'delete':
                $result = $statement->rowCount();
                break;
            default:
                $result = $status;
        }
        $time_end = microtime(true);

        $meta = $this->getStatementMeta($statement);

        $qr = new QueryResult($meta, $result, $this->queries);
        $qr->setResume([
            'type'       => $type,
            'rows'       => $statement->rowCount(),
            'elapsed_ms' => round(($time_end - $time_start) * 1000, 1)
        ]);

        return $qr;

    }

    private function getStatementMeta($statement)
    {
        $metas = new MetaCollect();

        $query = $statement->queryString;

        preg_match('/(FROM|INTO)\s([^\s]*)/i', $query, $match);

        $table = data_get($match, 2);

        // Récupération de la clé primaire de la table
        $pk_column_name = data_get($this->db->selectOne("EXEC sp_fkeys @pktable_name = '${table}'"), 'PKCOLUMN_NAME');

        for ($i = 0; $i < $statement->columnCount(); $i++) {
            $sm = $statement->getColumnMeta($i);

            $key = '';

            if ($sm['name'] == $pk_column_name) {
                $key = 'primary_key';
            }

            $meta = new Meta;
            $meta->setTable($table);
            $meta->setField($sm['name']);
            $meta->setType($sm['native_type']);
            //$meta->setNull(in_array('not_null', $sm['flags']) ? false : true);
            $meta->setKey($key);

            $metas->push($meta);
        }

        return $metas;
    }

    public function getQueryString($query)
    {
        return data_get($this->queries, $query);
    }


}
