<?php

namespace Goldenscarab\Modulus\Debusca\App\Models;

use Illuminate\Database\Eloquent\Model;
use Illuminate\Support\Facades\Session;
use Symfony\Component\Process\Exception\RuntimeException;

class Shell extends Model
{
    private $type;

	private $cout;

	private $alias = [];

    private $excludes = ['nano', 'vim', 'tail -f'];

    private $pwd;


    public function __construct($path)
    {
        if (!Session::has('shell_history_command')) {
            Session::push('shell_history_command', "");
        }

        if (!Session::has('shell_history_position')) {
            Session::put('shell_history_position', 0);
        }

        if (!Session::has('shell_pwd')) {
            Session::put('shell_pwd', $path);
            $this->pwd = $path;
        } else {
            $this->pwd = Session::get('shell_pwd');
        }

        $this->alias = [
            'll'   => 'ls -alF',
            'la'   => 'ls -A',
            'l'    => 'ls -CF',
            'cd ~' => 'cd ' . base_path()
        ];
    }

    /**
	 * Execute une commande Shell
	 * @param  string $cmd La commande à exécuter
	 * @return string      Le retour du terminal
	 */
	public function command($cmd, $error_off = false)
	{
		$cmd = trim($cmd);

		$cout  = '';

		$this->addCommandToHistory($cmd);

		// Gestion des alias
		$cmd = isset($this->alias[$cmd]) ? $this->alias[$cmd] : $cmd;

		//  En fonction de la commande
		if ($this->cmdIsExcluded($cmd)) {
			$cout .= 'sh: 1: ' . $cmd . ' is not a command allowed' . PHP_EOL;
		} else if ($this->cmdIsChangeDirectory($cmd)) {
			$cout .= $this->changeDirectory($cmd);
		} else if ($cmd == 'history'){
			$cout .= $this->renderHistory();
		} else {
			$cout .= $this->exec($cmd, $error_off);
		}

		$cout .= $this->renderPrompt();

		return $cout;
	}

    private function cmdIsExcluded($cmd)
	{
		foreach ($this->excludes as $exclude) {
			if (preg_match('/^' . $exclude . '.*/', $cmd)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Test si une commande est de type changement de dossier
	 * @param  string $cmd La commande à tester
	 * @return boolean     True si la commande et de type changement de dossier
	 */
	private function cmdIsChangeDirectory($cmd)
	{
		return preg_match("/^\s*cd\s+(.+)\s*$/", $cmd);
	}

	/**
	 * Changement de dossier
	 * @param  string $cmd La commande de changement de dossier
	 * @return void
	 */
	private function changeDirectory($cmd)
	{

        if ($this->isValidCommandResult($cmd)) {
            $this->pwd = trim($this->exec($cmd . " && pwd"));
            Session::put('shell_pwd', $this->pwd);
		} else {
            return $this->exec($cmd, true);
        }
	}

	/**
	 * Test si une commande retourne une erreur
	 * @param  string  $cmd La commande à tester
	 * @return boolean      True si la commande est valide
	 */
	private function isValidCommandResult($cmd)
	{
		$result = $this->exec($cmd, true);
		return preg_match("/^sh:\s1:.*/", $result) == 0;
	}

	private function addCommandToHistory($cmd)
	{
		Session::push('shell_history_command', $cmd);
        Session::put('shell_history_position', count(Session::get('shell_history_command')) -1);
	}

	public function getHistory($direction)
	{
		$history  = Session::get('shell_history_command');
		$position = Session::get('shell_history_position');

		if ($direction == 'up' && $position > 0) {
			$position--;
		} else if ($direction == 'down' && $position < count($history) - 1) {
			$position++;
		} else {
			return $this->renderPrompt();
		}

		$stdout = $this->renderPrompt() . $history[$position];

        Session::put('shell_history_position', $position);

		return $stdout;
	}

	/**
	 * Retourne le rendu de l'historique
	 * @return
	 */
	public function renderHistory()
	{
		$stdout = '';
		$template = '%s %s' . PHP_EOL;

		foreach (Session::get('shell_history_command') as $key => $history) {
			$stdout .= vsprintf($template, [$key, $history]);
		}

		return $stdout;
	}

	/**
	 * Demande de suggestion
	 * @param  string $command La commande source necessitant de l'aide
	 * @return string          Le retour de la console
	 */
	public function hint($command)
	{
		$cmd_split = explode(' ', $command);

		// Type de la suggestion demandée
		if (count($cmd_split) > 1) {
			if($cmd_split[0] == 'cd') {
				$type = 'dir';
			} else {
				$type = 'file';
			}
		} else {
			$type = 'cmd';
		}

		$search = array_pop($cmd_split);

		// Préparationd de la commande
		// if ($type == 'cmd') {
		// 	$cmd = "compgen -c " . $search;
		// } else if (){
		// 	$cmd = "compgen -f " . $search;
		// }

		switch ($type) {
			case 'cmd':
				$cmd = "compgen -c " . $search;
				break;
			case 'dir':
				$cmd = "compgen -d " . $search;
				break;

			default:
				$cmd = "compgen -f " . $search;
				break;
		}

		$cmd = "/bin/bash -c '" . $cmd . "'";

    	$hints = explode("\n", $this->exec($cmd));
    	$hints = array_filter($hints);

    	$stdout = '';

    	if (empty($hints)) {
    		$stdout =  null;
    	} else if (count($hints) > 1) {
    		$stdout = "\n";
    		$stdout .= implode("\n", $hints);
    		$stdout .= "\n";
    		$stdout .= $this->renderPrompt();
    		//$stdout .= trim($command);
    		$stdout .= $this->autocomplete($command, $hints);

    	} else {
    		//$stdout = str_replace($search, '', $hints[0]);
    		$stdout = $this->autocomplete($command, $hints);
    	}

    	return $stdout;
	}

	public function autocomplete($command, array $hints)
	{
		$cmd_split = explode(' ', $command);
		$search = array_pop($cmd_split);

		$autocomplete = $search;
		$max_size = 0;

		// Recherche de la chaine la plus longue
		foreach ($hints as $hint) {
			$size = strlen($hint);
			if ($size > $max_size) {
				$max_size = $size;
			}
		}

		if (count($hints) > 1) {

			$first_hint = array_shift($hints);

			// Comparaison des chaines de suggestion
			for ($i = 1 ; $i <= $max_size; $i++) {
				$identique = true;

				foreach ($hints as $hint)  {
					if(strncmp($first_hint, $hint, $i) != 0) {
						$identique = false;
					}
				}

				if ($identique) {
					$complete = substr($first_hint, 0, $i);
					$autocomplete = $command . str_replace($search, '', $complete);
				}
			}
		} else {
			$complete = $hints[0];
			$autocomplete = str_replace($search, '', $complete);
		}

		// S'il s'agit d'un cd et que le dossier est unique on ajout un /
		if (count($cmd_split) > 0 && $cmd_split[0] == 'cd' && strlen($complete) == $max_size) {
			$autocomplete .= '/';
		}

		return $autocomplete;
	}

	/**
	 * Affichage d'un texte de présentation
	 * @return string
	 */
	public function getPresentation()
	{
		$cout = '';
		$cout .= " _____                   _             _ \n";
		$cout .= "|_   _|__ _ __ _ __ ___ (_)_ __   __ _| |\n";
		$cout .= "  | |/ _ \ '__| '_ ` _ \| | '_ \ / _` | |\n";
		$cout .= "  | |  __/ |  | | | | | | | | | | (_| | |\n";
		$cout .= "  |_|\___|_|  |_| |_| |_|_|_| |_|\__,_|_|\n";
		$cout .= "\n";
		$cout .= php_uname() . "\n\n";
		$cout .= $this->renderPrompt();

		return $cout;
	}

	/**
	 * Retourne le prompt [user@hostname:path$]
	 * @return string
	 */
	public function renderPrompt()
	{
		$user         = get_current_user();
		$hostname     = gethostname();
        $pwd          = '';

		$folders = explode('/', $this->pwd);
		if (count($folders) > 2) {
			$folders = array_slice($folders, -2, 2);
			$path    = array_merge(['..'], $folders);
			$pwd     = implode('/', $path);
		}

		$template = '%s@%s:%s$ ';

		$prompt = vsprintf($template, [
			$user,
			$hostname,
			$pwd
		]);

		return $prompt;
	}

	/**
	 * Execution d'une commande en mode Shell
	 * @param  string $cmd [description]
	 * @return string      [description]
	 */
	public function exec($cmd, $error_off = false)
	{
        $template = debusca_path('bin/wrapper') . ' "%s" "%s"';
		$token = "O1s9dkFteyZqMFozP0xYY19QbzE0UXUtQmYsdCh7NTlNNzVSODZpcFQsbUY5TkBjUXY3MyttOCY4cFM3Rj1Idg==";

        $escaped = [
            '$' => "\\$"
        ];

        $cmd = str_replace(array_keys($escaped), array_values($escaped), $cmd);

		$cmd = 'cd \"' . $this->pwd . '\" && ' . $cmd . "    2>&1";

		$exec = vsprintf($template, [
			base64_decode($token),
			$cmd
		]);

		$stdout = shell_exec($exec);

        if (!$error_off) {
            if (preg_match('/Access error/', $stdout) === 1 || preg_match('/^([^: ]*:)\s.*/', $stdout) === 1) {
                throw new RuntimeException($stdout);
            }
        }

		return $stdout;
	}

}
